import types
from pyf_utils import PYF

WORD=int
DWORD=int
FLOAT=float
STRING=str
array=0

header = ["Header", ["major",WORD], ["minor",WORD], ["flags",DWORD]]
vector = ["Vector", ["x",FLOAT], ["y",FLOAT], ["z",FLOAT]]
coords2d = ["Coords2d", ["u",FLOAT], ["v",FLOAT]]
quaternion = ["Quaternion", ["s",FLOAT], ["v",vector]]
matrix4x4 = ["Matrix4x4", ["matrix",array,FLOAT,16]]
colorRGBA = ["ColorRGBA", ["red",FLOAT], ["green",FLOAT], ["blue",FLOAT],
	["alpha",FLOAT]]
colorRGB = ["ColorRGB", ["red",FLOAT], ["green",FLOAT], ["blue",FLOAT]]
indexed_colour = ["IndexedColor", ["index",DWORD], ["indexColor",colorRGBA]]
boolean = ["Boolean", ["truefalse",DWORD]]
boolean2d = ["Boolean2d", ["u",boolean], ["v",boolean]]
material = ["Material", ["faceColor",colorRGBA], ["power",FLOAT],
	["specularColor",colorRGB], ["emissiveColor",colorRGB]]
texture_filename = ["TextureFilename", ["filename",STRING]]
mesh_face = ["MeshFace", ["nFaceVertexIndices",DWORD],
	["faceVertexIndices",array,DWORD,"nFaceVertexIndices"]]
mesh_face_wraps = ["MeshFaceWraps", ["nFaceWrapValues",DWORD],
	["faceWrapValues",array,boolean2d,"nFaceWrapValues"]]
mesh_texture_coords = ["MeshTextureCoords", ["nTextureCoords",DWORD],
	["textureCoords",array,coords2d,"nTextureCoords"]]
mesh_normals = ["MeshNormals", ["nNormals",DWORD],
	["normals",array,vector,"nNormals"], ["nFaceNormals",DWORD],
	["faceNormals",array,mesh_face,"nFaceNormals"]]
mesh_vertex_colors = ["MeshVertexColors", ["nVertexColors",DWORD],
	["vertexColors",array,indexed_colour,"nVertexColors"]]
mesh_material_list = ["MeshMaterialList", ["nMaterials",DWORD],
	["nFaceIndexes",DWORD], ["faceIndexes",array,DWORD,"nFaceIndexes"],
	["materials",array,material,"nMaterials"]]
mesh = ["Mesh", ["nVertices",DWORD], ["vertices",array,vector,"nVertices"],
	["nFaces",DWORD], ["faces",array,mesh_face,"nFaces"]]
frame_transform_matrix = ["FrameTransformMatrix", ["frameMatrix",matrix4x4]]
frame = ["Frame"]
float_keys = ["FloatKeys",["nValues",DWORD], ["values",array,FLOAT,"nValues"]]
timed_float_keys = ["TimedFloatKeys", ["time",DWORD], ["tfkeys",float_keys]]
animation_key = ["AnimationKey", ["keyType",DWORD], ["nKeys",DWORD],
	["keys",array,timed_float_keys,"nKeys"]]
animation_options = ["AnimationOptions", ["openclosed",DWORD],
	["positionquality",DWORD]]
animation = ["Animation"]
animation_set = ["AnimationSet"]
	
objs = [header,vector,coords2d,quaternion,matrix4x4,colorRGBA,colorRGB,
	indexed_colour,boolean,boolean2d,material,texture_filename,
	mesh_face,mesh_face_wraps,mesh_texture_coords,mesh_normals,
	mesh_vertex_colors,mesh_material_list,mesh,frame_transform_matrix,
	frame,float_keys,timed_float_keys,animation_key,animation_options,
	animation,animation_set]

obj_dict = {}
map(lambda a:obj_dict.__setitem__(a[0],a), objs)

class DirectXObj:
	def __init__(this, type):
		this.type = type
		this.children = []
	def __repr__(this):
		s = this.type
		try:
			s += "(%s)" % (this.name,)
		except:
			pass
		if len(this.children) == 0:
			return s
		s += " ["
		for i in this.children:
			s += str(i) + ", "
		s = s[:-2]
		s += "]"
		return s

class DirectXTokenObj:
	def __init__(this, data):
		this.data = data[1:]
		this.current_line = data[0].strip()
	def next(this):
		if this.current_line == "":
			raise "StopIteration"
		next_rbrace = this.current_line.find("}")
		next_lbrace = this.current_line.find("{")
		if next_rbrace == -1:
			next_rbrace = len(this.current_line)
		if next_lbrace == -1:
			next_lbrace = len(this.current_line)
		if next_rbrace < next_lbrace:
			next_brace = next_rbrace
		else:
			next_brace = next_lbrace
		if (next_brace == -1):
			next_brace = len(this.current_line)
		next_scolon = this.current_line.find(";")
		if (next_scolon == -1):
			next_scolon = len(this.current_line)
		next_comma = this.current_line.find(",")
		if (next_comma == -1):
			next_comma = len(this.current_line)
		lsts = this.current_line.split()
		if next_brace == 0:
			token = this.current_line[0]
		elif next_scolon == 0:
			token = ";"
		elif next_comma == 0:
			token = ","
		elif len(lsts[0]) < next_brace and len(lsts[0]) < next_scolon \
			and len(lsts[0]) < next_comma:
			token = lsts[0]
		elif next_brace < next_scolon and next_brace < next_comma:
			token = lsts[0][:next_brace]
		elif next_comma < next_scolon:
			token = lsts[0][:next_comma]
		else:
			token = lsts[0][:next_scolon]
		this.current_line = this.current_line[len(token):].strip()
		if len(this.current_line) == 0:
			try:
				this.current_line = this.data[0].strip()
				while this.current_line[:2] == "//":
					this.data = this.data[1:]
					this.current_line = this.data[0].strip()
			except:
				this.current_line = ""
			this.data = this.data[1:]
		#print token
		return token
	def pushback(this, token):
		#print "->" + token
		this.current_line = token + " " + this.current_line

def directx_handle_guts(dtype_obj, data, dobj):
	next = data.next()
	#print "next is:", next
	if next == "{":
		ref = data.next()
		dobj.ref = ref
		assert data.next() == "}"
		data.pushback(";")
		return
	data.pushback(next)
	for i in dtype_obj[1:]:
		if i[1] == array:
			length = i[3]
			if type(length) == types.StringType:
				length = eval("dobj."+length)
			if type(i[2]) == types.ListType:
				arr = []
				for j in range(length):
					dobji = DirectXObj(i[2][0])
					directx_handle_guts(i[2], data, dobji)
					arr.append(dobji)
					dnext = data.next()
					assert dnext == "," or dnext == ";"
			else:
				arr = []
				cap_sem = 0
				for j in range(length):
					tdata = i[2](data.next())
					arr.append(tdata)
					dnext = data.next()
					if dnext == ",":
						cap_sem = 1
					assert dnext == ";" or dnext == ","
				if cap_sem:
					assert data.next() == ";"
			exec("dobj."+i[0] + "=arr")
		elif type(i[1]) == types.ListType:
			dobji = DirectXObj(i[1][0])
			directx_handle_guts(i[1], data, dobji)
			exec("dobj."+i[0] + "=dobji")
			assert data.next() == ";"
		else:
			tdata = i[1](data.next())
			exec("dobj."+i[0] + "=tdata")
			#print "dobj."+i[0] + "=tdata"
			assert data.next() == ";"

def directx_handle_object(data):
	dtype = data.next()
	dobj = DirectXObj(dtype)
	
	name = data.next()
	if not name == "{":
		dobj.name = name
		assert data.next() == "{"
	
	dtype_obj = obj_dict[dtype]
	directx_handle_guts(dtype_obj, data, dobj)
	
	next = data.next()
	while not next == "}":
		data.pushback(next)
		child = directx_handle_object(data)
		dobj.children.append(child)
		assert data.next() == "}"
		next = data.next()
	data.pushback(next)
	return dobj

def directx_to_tree(data):
	objects = []
	while 1:
		objects.append(directx_handle_object(data))
		assert data.next() == "}"
		try:
			data.pushback(data.next())
		except:
			return objects


def directx_mesh_normals(n_obj):
	# first, grab the normals
	normals = map(lambda a: (a.x, a.y, a.x), n_obj.normals)
	# The normals face list should match the face list.
	# So if I create a list of normals based around the order in
	# the normals face list, then I should have a correct list.
	normal_list = []
	for face in n_obj.faceNormals:
		for index in face.faceVertexIndices:
			normal_list.append(normals[index])
	return normal_list

def directx_fix(name):
	dot = name.find(".")
	while dot > -1:
		name = name[:dot]+name[dot+1:]
		dot = name.find(".")
	if name == "Material":
		name = "_Material"
	return name

def directx_tree_to_code(tree, out_file, name):
	pyf = PYF(out_file)
	pyf.header(name)
	#pyf.container(name + "_C")

	index = 0
	for elt in tree:
		if elt.type == "Header":
			continue
		elif elt.type == "Material":
			amb = elt.faceColor
			em = elt.emissiveColor
			spec = elt.specularColor
			shin = elt.power
			amb = (amb.red, amb.blue, amb.green, amb.alpha)
			em = (em.red, em.blue, em.green, 1)
			spec = (spec.red, spec.blue, spec.green, 1)
			mname = directx_fix(elt.name)
			pyf.material(mname, amb, em, spec, shin)
		elif elt.type == "Mesh":
			# vertices. map to double tuples
			vertices = map(lambda a: (a.x, a.y, a.z), elt.vertices)
			# generate VectorList
			try:
				vname = directx_fix(elt.name)
			except:
				vname = "vl" + str(index)
				index += 1
			pyf.vectorList(vname, vertices)

			pyf.container(vname+"_C")

			# faces.  split quads and tris
			qlist = []
			tlist = []
			for face in elt.faces:
				if face.nFaceVertexIndices == 4:
					qlist.append(face)
				else:
					tlist.append(face)
			
			# map to single list
			qlist = reduce(lambda a, b: a + b.faceVertexIndices,
				qlist, [])
			
			tlist = reduce(lambda a,b: a + b.faceVertexIndices,
				tlist, [])

			use_normals = 2
			for child in elt.children:
				if child.type == "MeshMaterialList":
					# for now, just use first vertex
					# colour
					col = child.faceIndexes[0]
					mat = child.materials[col]
					pyf.setMaterial(vname + "_C", 
						directx_fix(mat.ref))
				if child.type == "MeshNormals":
					nlist = directx_mesh_normals(child)
					pyf.vectorList(vname+"_N", nlist)
					use_normals = 1

			if len(qlist) > 0:
				pyf.intList(vname + "_quads", qlist)
				if use_normals:
					pyf.staticRender(
						vname + "_qr",
						vname + "_C",
						vname, 
						"GL_QUADS",
						vname + "_quads", 
						vname+"_N")
				else:
					pyf.staticRender(
						vname + "_qr",
						vname + "_C",
						vname, 
						"GL_QUADS",
						vname + "_quads")
			if len(tlist) > 0:
				pyf.intList(vname + "_tris", tlist)
				if use_normals:
					pyf.staticRender(
						vname + "_tr",
						vname + "_C",
						vname, 
						"GL_TRIANGLES",
						vname + "_tris", 
						vname+"_N")
				else:
					pyf.staticRender(
						vname + "_tr",
						vname + "_C",
						vname, 
						"GL_TRIANGLES",
						vname + "_tris")
	pyf.finish()				

def directx_to_code(in_file, out_file, name):
	lines = in_file.readlines()

	#check header line
	header = lines[0]
	lines = lines[1:]
	assert header[:4] == "xof "
	version_major = int(header[4:6])
	version_minor = int(header[6:8])
	assert version_major == 3
	assert version_minor == 2
	
	# can only handle txt for now
	assert header[8:12] == "txt "

	float_size = int(header[12:16])

	tree = directx_to_tree(DirectXTokenObj(lines[1:]))
	#print tree
	directx_tree_to_code(tree, out_file, name)

import sys
directx_to_code(open(sys.argv[1], "r"), open("foo.py","w"), "foo")
