/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyIntListType;

static void C_PyIntList_Append(PyObject *obj, int val);
static int C_PyIntList_Item(PyObject *self, int index, int *store);
static int *PyIntList_Ints(PyObject *self);
static int PyIntList_Length(PyObject *self);
static IIntList PyIntList_IIntList = {C_PyIntList_Append, 
                C_PyIntList_Item, PyIntList_Ints, PyIntList_Length};
static IC PyIntList_Interfaces = {NULL, &PyIntList_IIntList};

int PyIntList_Check(PyObject *v) {

        return v->ob_type == &PyIntListType;

}

static int * PyIntList_Ints(PyObject *v) {

        return ((PyIntList *)v)->ints;

}

static int PyIntList_Length(PyObject *v) {

        return ((PyIntList *)v)->length;

}

static inline int PyIntList_Alloced(PyObject *v) {

        return ((PyIntList *)v)->alloced;

}

PyObject *C_PyIntList_New(void) {

        PyIntList *l = PyObject_New(PyIntList, &PyIntListType);

        D_ALLOC(l, "IntList");

        l->alloced = l->length = 0;
        l->interfaces = &PyIntList_Interfaces;
        l->ints = NULL;

        return (PyObject *)l;
        
}

PyObject *C_PyIntList_OfLength(int length) {

        PyIntList *l = (PyIntList *)C_PyIntList_New();
        l->alloced = length;
        
        if (l->alloced > 0) {
                l->ints = malloc(sizeof(int) * l->alloced);
        }

        return (PyObject *)l;

}
        
PyObject *PyIntList_New(PyObject *self, PyObject *args) {

        PyObject *l;
        int size;

        if (PyTuple_Size(args) == 1) {

                l = PyTuple_GetItem(args, 0);

                if (!PyInt_Check(l)) {
                        return NULL;
                }

                size = (int)PyInt_AsLong(l);

                return (PyObject *)C_PyIntList_OfLength(size);

        } else if (PyTuple_Size(args) == 0) {

                return (PyObject *)C_PyIntList_New();

        }

        return NULL;
        
}

static void C_PyIntList_Free(PyObject *obj) {

        PyIntList *l = (PyIntList *)obj;

        D_FREE(l, "IntList");
        
        if (l->alloced > 0) {
                free(l->ints);
        }

        PyObject_Del((PyObject *)obj);
}

static void PyIntList_Free(PyObject *self) {

        C_PyIntList_Free((void *)self);
        
}

static int C_PyIntList_Item(PyObject *self, int index, int*store) {
	
	PyIntList *l = (PyIntList *)self;

	if (index > l->length) {
		
		return -1;
		
	}
	
	*store = l->ints[index];
	return 0;
	
}

static PyObject *PyIntList_Item(PyObject *self, int index) {
	
	int val;
	
	if (C_PyIntList_Item(self, index, &val) == -1) {
		return NULL;
	}
	
	return PyInt_FromLong(val);
	
}
	

static void C_PyIntList_Append(PyObject *obj, int val) {

        PyIntList *l = (PyIntList *)obj;

        if (l->alloced == l->length) {
                
                if (l->alloced == 0) {
                        l->alloced++;
                } else {
                        l->alloced *= 2;
                }

                l->ints = realloc(l->ints, l->alloced * sizeof(int));

        }

        
        l->ints[l->length++] = val;


}

static PyObject *PyIntList_Append(PyObject *self, PyObject *args) {

        int val;
        
        if (!PyArg_ParseTuple(args, "i", &val)) {
                return NULL;
        }

        C_PyIntList_Append(self, val);

        Py_INCREF(Py_None);
        return Py_None;

}

static PyMethodDef PyIntList_methods[] = {
        {"append", (PyCFunction)PyIntList_Append, METH_VARARGS,
                "Appends an integer to the int list"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyIntList_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyIntList_methods, self, name);

}
        
static PySequenceMethods PyIntList_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        PyIntList_Item, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyIntListType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyIntList",
        sizeof(PyIntList),
        0,
        PyIntList_Free,
        0, /* print */
        PyIntList_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyIntList_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
