/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

/* empty comment */

staticforward PyTypeObject PyRenderType;

static void C_PyRender_Render(PyObject *self, int time, int picking);
static int C_PyRender_Compile(PyObject *self) {return FALSE;}

static IRender PyRender_IRender = {C_PyRender_Render, C_PyRender_Compile};
static IC PyRender_Interfaces = {&PyRender_IRender, NULL};

static inline int PyRender_Check(PyObject *v) {

        return v->ob_type == &PyRenderType;

}

static inline PyObject * PyRender_Vectors(PyObject *v) {

        return ((PyRender *)v)->vectors;

}

static inline PyObject * PyRender_Normals(PyObject *v) {

        return ((PyRender *)v)->normals;

}

static inline int PyRender_Length(PyObject *v) {

        return ((PyRender *)v)->length;

}

PyObject *C_PyRender_New(PyObject *vectors, int mode, PyObject *normals, 
                PyObject *faces, int length) {


        PyRender *r = PyObject_New(PyRender, &PyRenderType);
        r->interfaces = &PyRender_Interfaces;
        r->vectors = vectors;
        r->mode = mode;
        r->normals = normals;
        r->vertexlist = faces;
        r->length = length;

        Py_INCREF(vectors); 
        if (normals != NULL) {
                Py_INCREF(normals);
        }
        if (faces != NULL) {
                Py_INCREF(faces);
        }

        return (PyObject *)r;
        
}

PyObject *PyRender_New(PyObject *self, PyObject *args, PyObject *kwords) {

        PyObject *vectors, *normals = NULL, *is_static = NULL, *faces = NULL;
        int length, is_s = 0;
        int mode = GL_POINTS;
					
        static char *kwlist[] = {"vectors", "mode", "normals", 
                "faces", NULL};
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "Oi|OO",
                kwlist, &vectors, &mode, &normals, &faces)) {
                return NULL;
        }

        /* FIXME: These should eventually use the C interfaces to get at this
         * data */

        length = VL_LENGTH(vectors);
        if (normals != NULL) {
                int l2 = VL_LENGTH(normals);
                if (l2 < length) {
                        length = l2;
                }
        }
        
        if (is_static != NULL && PyInt_Check(is_static)) {
                is_s = (int)PyInt_AsLong(is_static);
        }

        return (PyObject *)C_PyRender_New(vectors, mode, normals, faces,
                        length);

}
        
static void PyRender_Free(PyObject *self) {
        
}

static void C_Render_Lists(double *v, double *n, int *faces, int length, 
                int type) {

        
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, GL_DOUBLE, 0, v);
        if (n != NULL) {
                glEnableClientState(GL_NORMAL_ARRAY);
                glNormalPointer(GL_DOUBLE, 0, n);
        } else {
                glDisableClientState(GL_NORMAL_ARRAY);
        }

        if (faces == NULL) {
                glDrawArrays(type, 0, length);
        } else {
                glDrawElements(type, length, GL_UNSIGNED_INT, faces);
        }

        
}

static int PyRender_Setattr(PyObject *obj, char *name, PyObject *v) {
	
	PyRender *r = (PyRender *)obj;
	
	if (strcmp(name, "mode") == 0) {
		
		r->mode = (int)PyInt_AsLong(v);
		
		return 0;
		
	}
	
	return -1;
}

static PyObject *PyRender_Getattr(PyObject *obj, char *name) {
	
	return NULL;
	
}

static void C_PyRender_Render(PyObject *self, int time, int picking) {

        PyRender *r = (PyRender *)self;
        double *v, *n = NULL;
        int *f = NULL;
        
        glPushMatrix();

        v = VL_VALUES(r->vectors);
        
        if (r->normals != NULL) {
                n = VL_VALUES(r->normals);
        }

        if (r->vertexlist != NULL) {
                f = IL_INTS(r->vertexlist);
        }
        
        if (f == NULL) {
                C_Render_Lists(v, n, f, r->length, r->mode);
        } else {
                C_Render_Lists(v, n, f, IL_LENGTH(r->vertexlist), 
                                r->mode);
        }
		
		
        glPopMatrix();
        
}

static PyObject *PyRender_Call(PyObject *self, PyObject *args, PyObject *dict) {

        PyObject *p;
        int gl_type = GL_POINTS;


        if (!(PyTuple_Size(args) == 1)) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);

        if (p != Py_None) {

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_type = (int) PyInt_AsLong(p);

        }

        C_PyRender_Render((void *)self, 0, 0);

        Py_INCREF(Py_None);
        return Py_None;

}
        
static PySequenceMethods PyRender_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        0, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyRenderType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyRender",
        sizeof(PyRender),
        0,
        PyRender_Free,
        0, /* print */
        PyRender_Getattr, /* getattr */
        PyRender_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyRender_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyRender_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
