/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

staticforward PyTypeObject PyRippleBehaviourType;

void C_PyRippleBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time);
static IDSBehaviour PyRippleBehaviour_IDSBehaviour 
	= {C_PyRippleBehaviour_Run};

static IC PyRippleBehaviour_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, &PyRippleBehaviour_IDSBehaviour};

void C_PyRippleBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time) {

	int i, j;
	PyRippleBehaviour *d = (PyRippleBehaviour *)self;
	
	for (i = 0; i < width; i++)
	{
		double xdiff, xdiffn;
		xdiff = sin(3.142*2*((float)i/width/d->xper + 
				time/d->xtimeper)*3.142);
		xdiffn = cos(3.142*2*((float)i/width/d->xper + 
				time/d->xtimeper)*3.142)/d->xper;
		for (j = 0; j < height; j++)
		{
			double ydiff, ydiffn;
			ydiff = sin(3.142*2*((float)j/height/d->yper + 
					time/d->ytimeper)*3.142);
			ydiffn = cos(3.142*2*((float)j/height/d->yper + 
					time/d->ytimeper)*3.142)/d->yper;
			v[2] += ydiff + xdiff;
			v += 3;
			n[0] = xdiffn;
			n[1] = ydiffn;
			n += 3;
		}
	}
}

PyObject *C_PyRippleBehaviour_New(double xper, double yper, double xtimeper,
		double ytimeper) {
	
	PyRippleBehaviour *d = PyObject_New(PyRippleBehaviour,
			&PyRippleBehaviourType);


	d->interfaces = &PyRippleBehaviour_Interfaces;
	d->xper = xper;
	d->yper = yper;
	d->xtimeper = xtimeper;
	d->ytimeper = ytimeper;
	
	return (PyObject *)d;
	

}

PyObject *PyRippleBehaviour_New(PyObject *self, PyObject *args) {

	double xper = 1, yper = 1, xtimeper = 1000, ytimeper = 1000;
	
	if (!PyArg_ParseTuple(args, "|dddd", &xper, &yper, &xtimeper, 
				&ytimeper)) {
		return NULL;
	}

	return C_PyRippleBehaviour_New(xper, yper, xtimeper, ytimeper);
	
}

static int PyRippleBehaviour_Setattr(PyObject *obj, char *name,
		PyObject *val) {

	/*
	PyRippleBehaviour *data = (PyRippleBehaviour *)obj;
	
	if (strcmp(name, "xPeriodPeriod") == 0)
	{
		data->xperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodPeriod") == 0)
	{
		data->yperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "xPeriodMagnitude") == 0)
	{
		data->xpermag = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodMagnitude") == 0)
	{
		data->ypermag = PyFloat_AsDouble(val);
	}
	else
	{
		return -1;
	}
	*/

	return 0;
}	

static void PyRippleBehaviour_Free(PyObject *this) {

}

static PyTypeObject PyRippleBehaviourType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyRippleBehaviour",
        sizeof(PyRippleBehaviour),
        0,
        PyRippleBehaviour_Free,
        0, /* print */
        0, /* getattr */
        PyRippleBehaviour_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
