/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

#include "Vector_Privates.h"

#include <pthread.h>

/* empty comment */

staticforward PyTypeObject PyStaticRenderType;

static void C_PyStaticRender_Render(PyObject *self, int type, int picking);
static int C_PyStaticRender_Compile(PyObject *self);

static IRender PyStaticRender_IRender = {C_PyStaticRender_Render, 
        C_PyStaticRender_Compile};

static IC PyStaticRender_Interfaces = {&PyStaticRender_IRender, NULL};

static inline int PyStaticRender_Check(PyObject *v) {

        return v->ob_type == &PyStaticRenderType;

}

static inline PyObject * PyStaticRender_Vectors(PyObject *v) {

        return ((PyStaticRender *)v)->vectors;

}

static inline PyObject * PyStaticRender_Normals(PyObject *v) {

        return ((PyStaticRender *)v)->normals;

}

static inline int PyStaticRender_Length(PyObject *v) {

        return ((PyStaticRender *)v)->length;

}

void C_PyStaticRender_Quads_GenNormals(PyObject *vectors, 
                PyObject **normals, PyObject **faces) {

        /*
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        */
        /* unfinished ! */
}

void C_PyStaticRender_QuadStrip_GenNormals(PyObject *vectors, 
                PyObject **normals, PyObject **faces) {
        
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        int i;

        for (i = 0; i + 2 < v->length; i += 2)
        {
                PyVector *v0 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 1],
                                v->pyVectors[i]);
                PyVector *v1 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 3],
                                v->pyVectors[i + 1]);

                VL_APPEND((PyObject *)n, C_PyVector_Cross((PyObject *)v0, 
                                                (PyObject *)v1));
        }

        *normals = (PyObject *)n;

}

void hello_world(void *arg) {
	while (1) {
		printf("Hello world\n");
		sleep(1);
	}
}



void *C_PyStaticRender_New(PyObject *vectors, int mode, PyObject *normals, 
                PyObject *faces, PyObject *textureCoords, int textureMode, 
                int length, int normalMode, int genNormals) {


        PyStaticRender *r = PyObject_New(PyStaticRender, &PyStaticRenderType);

        D_ALLOC(r, "StaticRender");

        r->interfaces = &PyStaticRender_Interfaces;
        r->vectors = vectors;
        r->mode = mode;
        r->normals = normals;
        r->vertexlist = faces;
        r->length = length;
        r->normal_mode = normalMode;
        r->display_list = 0;
        r->textureCoords = textureCoords;
        r->textureMode = textureMode;

        Py_INCREF(vectors);
        if (normals != NULL)
                Py_INCREF(normals);
        if (faces != NULL)
                Py_INCREF(faces);

        if (textureCoords != NULL)
                Py_INCREF(textureCoords);
        
        if (normals != NULL || (!genNormals) || normalMode == 0)
                return (void *)r;

        /*
         * Not all possibilities implemented
         */
        if (faces == NULL) {
                
                switch (mode) {
                case GL_QUAD_STRIP:
                        C_PyStaticRender_QuadStrip_GenNormals(vectors, 
                                        &(r->normals), &(r->vertexlist));
                        break;
                case GL_QUADS:
                        C_PyStaticRender_Quads_GenNormals(vectors,
                                        &(r->normals), &(r->vertexlist));
                        break;

                }

        }

        return (void *)r;
        
}

PyObject *PyStaticRender_New(PyObject *self, PyObject *args, PyObject *kwords) {

        PyObject *vectors, *normals = NULL, *faces = NULL;
        PyObject *textureCoords = NULL;
        int length;
        int textureMode = 0;
        int mode = GL_POINTS;
        int gen_normals = 0;
        int normalMode = 0;
					
        static char *kwlist[] = {"vectors", "mode", "normals", 
                "faces", "normalMode", "genNormals", "textureCoords",
                "textureMode", NULL};
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "Oi|OOiiOi",
                kwlist, &vectors, &mode, &normals, &faces, &normalMode,
                &gen_normals, &textureCoords, &textureMode)) {
                return NULL;
        }

        /* FIXME: These should eventually use the C interfaces to get at this
         * data */

        length = VL_LENGTH(vectors);

        /* TODO: Implement check for length of normals */
        
        return (PyObject *)C_PyStaticRender_New(vectors, mode, normals, 
                        faces, textureCoords, textureMode, length, normalMode, 
                        gen_normals);
}


static void PyStaticRender_Free(PyObject *self) {

        D_FREE(self, "StaticRender");
        
        /* TODO */
        
}

void C_PyStaticRender_SetTextureCoords(PyObject *obj, PyObject *v) {

        PyStaticRender *r = (PyStaticRender *)obj;
        
        if (r->textureCoords != NULL)
        {
                Py_DECREF(r->textureCoords);
        }
        
        Py_INCREF(v);
        r->textureCoords = v;


}

static int PyStaticRender_Setattr(PyObject *obj, char *name, PyObject *v) {
	
	PyStaticRender *r = (PyStaticRender *)obj;
	
	if (strcmp(name, "mode") == 0) {
		
		r->mode = (int)PyInt_AsLong(v);
		
		return 0;
		
        }

        if (strcmp(name, "textureCoords") == 0) {

                C_PyStaticRender_SetTextureCoords(obj, v);
                return 0;
                  
        }
	
	return -1;
}

static PyObject *PyStaticRender_Getattr(PyObject *obj, char *name) {
	
	PyStaticRender *r = (PyStaticRender *)obj;

        if (strcmp(name, "textureCoords") == 0) {

                if (r->textureCoords == NULL) {
                        Py_INCREF(Py_None);
                        return Py_None;
                }

                Py_INCREF(r->textureCoords);
                return r->textureCoords;

        }
        
	return NULL;
	
}

inline void _texture(int mode, float *texture, int pos)
{

        if (texture == NULL)
                return;

        if (mode == 0) /* 2D Textures */
        {
                glTexCoord2f(texture[pos*2], texture[pos*2+1]);
        }
}
                

static void C_StaticRender_Lists_Static(PyObject *self, double *v, double *n, 
		int *faces, float *texture, int length, int type, 
		int normalMode, int tMode) {

        int i;

        if (normalMode == 1)
        {

                int n_pos = 0, v_pos = 0;
                
                assert(n != NULL);
                

                switch (type) {
                        case GL_QUAD_STRIP:
                                
                                glBegin(GL_QUADS);

                                while (v_pos + 4 <= length) {

                                        glNormal3dv(&n[(n_pos++)*3]);
                                        _texture(tMode, texture, v_pos);
                                        glVertex3dv(&v[(v_pos)*3]);
                                        _texture(tMode, texture, v_pos + 1);
                                        glVertex3dv(&v[(v_pos+1)*3]);
                                        _texture(tMode, texture, v_pos + 3);
                                        glVertex3dv(&v[(v_pos+3)*3]);
                                        _texture(tMode, texture, v_pos + 2);
                                        glVertex3dv(&v[(v_pos+2)*3]);

                                        v_pos += 2;
                                }

                                glEnd();
                                
                                break;

                default:
                        assert(!"Per-face normals for this render mode not yet"
                                " implemented!");
                }


                return;
                
        }

        
        glBegin(type);
	
        if (faces == NULL) {
        
                for (i = 0; i < length; i++) {
                
                        if (n != NULL) {
                                glNormal3dv(&n[i*3]);
                        }

                        _texture(tMode, texture, i);
                        glVertex3dv(&v[i*3]);

                }
                
        } else {
                for (i = 0; i < length; i++) {
                        if (n != NULL) {
                                glNormal3dv(&n[faces[i] * 3]);
                        }
                        _texture(tMode, texture, faces[i]);
                        glVertex3dv(&v[faces[i] * 3]);
                }
        }


        glEnd();
                        
}

static int C_PyStaticRender_Compile(PyObject *self) {

        double *v, *n = NULL;
        int *f = NULL;
        float *t = NULL;
        int length;
	
        PyStaticRender *r = (PyStaticRender *)self;

	TPRINTF("Compiling %p\n", self);
        if (r->display_list == 0)
        {
        
                v = VL_VALUES(r->vectors);
                if (r->normals != NULL) {
                        n = VL_VALUES(r->normals);
                }

                if (r->vertexlist != NULL) {
                        f = IL_INTS(r->vertexlist);
			length = IL_LENGTH(r->vertexlist);
                }
		else
		{
			length = r->length;
		}
                
                r->display_list = glGenLists(1);

                if (r->textureCoords != NULL) {
                        t = ((PyFloatList *)r->textureCoords)->floats;
                }
                
                if (r->display_list == 0)
                {
                        /* FIXME: Is this the correct behaviour? */
                        return FALSE;
                }
                glNewList(r->display_list, GL_COMPILE);

                C_StaticRender_Lists_Static(self, v, n, f, t, length, r->mode,
                                r->normal_mode, r->textureMode);
                glEndList();
        }

        return TRUE;
                
}

static void C_PyStaticRender_Render(PyObject *self, int time, int picking) {

        PyStaticRender *r = (PyStaticRender *)self;

        if (r->display_list == 0) {
		TPRINTF("Compiling %p\n", self);
                C_PyStaticRender_Compile(self);
        }
	TPRINTF("Calling %p\n", r);

	//printf("Writing name %p\n", self);
	//glLoadName((int)self);
        glCallList(r->display_list);
        
}

static PyObject *PyStaticRender_Call(PyObject *self, PyObject *args, PyObject *dict) {

        PyObject *p;
        int gl_type = GL_POINTS;


        if (!(PyTuple_Size(args) == 1)) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);

        if (p != Py_None) {

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_type = (int) PyInt_AsLong(p);

        }

        C_PyStaticRender_Render((void *)self, 0, 0);

        Py_INCREF(Py_None);
        return Py_None;

}
        
static PySequenceMethods PyStaticRender_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        0, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyStaticRenderType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyStaticRender",
        sizeof(PyStaticRender),
        0,
        PyStaticRender_Free,
        0, /* print */
        PyStaticRender_Getattr, /* getattr */
        PyStaticRender_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyStaticRender_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyStaticRender_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
