/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

PyObject *TograSphere(PyObject *self, PyObject *args);
PyObject *TograBox(PyObject *self, PyObject *args);
PyObject *TograCylinder(PyObject *self, PyObject *args);
PyObject *TograCircle(PyObject *self, PyObject *args);

PyObject *PyMatrix_NewTranslation(PyObject *self, PyObject *args);
PyObject *PyMatrix_NewRotation(PyObject *self, PyObject *args);
PyObject *PyMatrix_NewScale(PyObject *self, PyObject *args);

PyObject *PyTransformContainer_New(PyObject *self, PyObject *args);

PyObject *PyVectorList_Zip(PyObject *self, PyObject *args);
PyObject *TograExtrude(PyObject *self, PyObject *args);

PyObject *PyStaticTexture_New(PyObject *self, PyObject *args);
PyObject *PyGstTexture_New(PyObject *self, PyObject *args);

PyObject *PyLightObject_New(PyObject *self, PyObject *args, PyObject *kwords);

PyObject *PyStaticSurface_New(PyObject *self, PyObject *args, PyObject *kwords);
PyObject *PyDynamicSurface_New(PyObject *self, PyObject *args, 
		PyObject *kwords);

PyObject *PyWobbleBehaviour_New(PyObject *self, PyObject *args);
PyObject *PyRippleBehaviour_New(PyObject *self, PyObject *args);
PyObject *PyInterpolationBehaviour_New(PyObject *self, PyObject *args);

PyObject *PyMatrix_FromMatrix(PyObject *self, PyObject *args);

PyObject *PyODETransform_New(PyObject *self, PyObject *args);

PyObject *PyODEMixin_New(PyObject *self, PyObject *args);
PyObject *PyProtoMixin_New(PyObject *self, PyObject *args);
PyObject *PyProtoRender_New(PyObject *self, PyObject *args);

PyObject *PyGraphMixin_New(PyObject *self, PyObject *args);
PyObject *PyCustomGraphMixin_New(PyObject *self, PyObject *args);

PyObject *PyGraph_New(PyObject *self, PyObject *args);

PyMethodDef Togra_methods[] = {
        {"Vector", PyVector_New, METH_VARARGS, 
                "Create a new Vector"},
        {"VectorList", PyVectorList_New, METH_VARARGS, 
                "Create a new Vector list"},
        {"IntList", PyIntList_New, METH_VARARGS,
                "Create a new list of integers"},
        {"FloatList", PyFloatList_New, METH_VARARGS,
                "Create a new list of floats"},
        {"Render", (PyCFunction)PyRender_New, METH_VARARGS | METH_KEYWORDS,
                "Create a new Render object"},
        {"Light", (PyCFunction)PyLightObject_New, METH_VARARGS | METH_KEYWORDS,
                "Create a new light place-holder"},
        {"StaticRender", (PyCFunction)PyStaticRender_New, 
                METH_VARARGS | METH_KEYWORDS,
                "Create a new StaticRender object"},
        {"StaticSurface", (PyCFunction)PyStaticSurface_New, 
                METH_VARARGS | METH_KEYWORDS,
                "Create a new StaticSurface object"},
        {"DynamicSurface", (PyCFunction)PyDynamicSurface_New, 
                METH_VARARGS | METH_KEYWORDS,
                "Create a new DynamicSurface object"},
        {"Transform", (PyCFunction)PyTransform_New, 
                METH_VARARGS | METH_KEYWORDS,
                "Create a new Transform"},
	{"ODETransform", (PyCFunction)PyODETransform_New, METH_VARARGS,
		"Create a new ODE Transform"},
        {"InterpolatedPath", PyInterpolatedPath_New, METH_VARARGS,
                "Create a new InterpolatedPath object"},
        {"Container", PyContainer_New, METH_VARARGS,
                "Create a new Container object"},
        {"Material", (PyCFunction)PyMaterial_New, METH_VARARGS | METH_KEYWORDS,
                "Create a new material object"},
        {"Reflect", PyReflect_New, METH_VARARGS,
                "Create a new reflection object"},
        {"Translation", PyMatrix_NewTranslation, METH_VARARGS,
                "Create a new matrix with a specified translation"},
        {"Rotation", PyMatrix_NewRotation, METH_VARARGS,
                "Create a new matrix with a specified rotation"},
        {"Scale", PyMatrix_NewScale, METH_VARARGS,
                "Create a new matrix with a specified scale"},
	{"Matrix", PyMatrix_FromMatrix, METH_VARARGS,
		"Create a new matrix from a list of ints"},
        {"TransformContainer", PyTransformContainer_New, METH_VARARGS,
                "Create a new transform container"},
        {"StaticTexture", PyStaticTexture_New, METH_VARARGS,
                "Use an array of pixels as a fixed texture"},
        {"WobbleBehaviour", PyWobbleBehaviour_New, METH_VARARGS,
                "Example dynamic wobbly behaviour for DynamicSurface objects"},
        {"RippleBehaviour", PyRippleBehaviour_New, METH_VARARGS,
                "Example ripple behaviour for DynamicSurface objects"},
        {"InterpolationBehaviour", PyInterpolationBehaviour_New, METH_VARARGS,
                "Example interpolation behaviour for DynamicSurface objects"},
        {"GstTexture", PyGstTexture_New, METH_VARARGS,
                "Play a movie in a texture using GStreamer"},
        {"init", PyRenderControl_New, METH_VARARGS,
                "Start Togra and create a new RenderControl object"},
        {"sphere", TograSphere, METH_VARARGS,
                "Create a sphere of specified size"},
        {"box", TograBox, METH_VARARGS,
                "Create a box of specified dimensions"},
        {"cylinder", TograCylinder, METH_VARARGS,
                "Create a cylinder of specified dimensions"},
        {"circle", TograCircle, METH_VARARGS,
                "Create a circle of specified dimensions"},
        {"zip", PyVectorList_Zip, METH_VARARGS,
                "Zip two vectorlists together"},
        {"extrude", TograExtrude, METH_VARARGS,
                "Extrude a vectorList"},
	{"ODEMixin", PyODEMixin_New, METH_VARARGS,
		"Create a new ODE physics mixin"},
	{"ProtoMixin", PyProtoMixin_New, METH_VARARGS,
		"Create a new ODE physics mixin"},
	{"GraphMixin", PyGraphMixin_New, METH_VARARGS,
		"Create a new graph model mixin"},
	{"CustomGraphMixin", PyCustomGraphMixin_New, METH_VARARGS,
		"Create a new custom graph model mixin"},
	{"ProtoRender", PyProtoRender_New, METH_VARARGS,
		"Create a new ODE physics mixin"},
	{"Graph", PyGraph_New, METH_VARARGS,
		"Create a new graph object"},
        {NULL, NULL, 0, NULL}
};

DL_EXPORT(void) initTogra(void) {

		InitTime();
        Py_InitModule("Togra", Togra_methods);

}
