/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

/*
typedef struct {
        void (*generateTransform)(PyObject *self, PyObject *status, int time, 
                        int gen_forward, int gen_reverse, int apply);
        void (*transformByMe)(PyObject *self);
        void (*invertByMe)(PyObject *self);
        int (*registerSize)(PyObject *self);
        void (*registerFunc)(PyObject *self, PyObject *status, int delay,
                        PyObject *callback, PyObject *user, PyObject *user2);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
} ITransform;
*/
#include "TograIncludes.h"
staticforward PyTypeObject PyTransformType;


static void C_PyTransform_GenerateTransform(PyObject *self, PyObject *status,
                int time, int gen_forward, int gen_reverse, int apply);
inline void C_PyTransform_TransformByMe(PyObject *self);
inline void C_PyTransform_InvertByMe(PyObject *self);
static int C_PyTransform_RegisterSize(PyObject *self) {return 0;}

static ITransform PyTransform_ITransform = {C_PyTransform_GenerateTransform,
        C_PyTransform_TransformByMe, C_PyTransform_InvertByMe, 
        C_PyTransform_RegisterSize, NULL, NULL};

static IC PyTransform_Interfaces = {NULL, NULL, NULL, NULL, NULL, 
        &PyTransform_ITransform};

static inline int PyTransform_Check(PyObject *v) {

        return v->ob_type == &PyTransformType;

}

static inline PyObject * PyTransform_Translation(PyObject *v) {

        return ((PyTransform *)v)->translation;

}

static inline PyObject * PyTransform_Scale(PyObject *v) {

        return ((PyTransform *)v)->scale;

}

PyObject *C_PyTransform_New(PyObject *translate, PyObject *rotateAxis, 
	double rotateAmount, PyObject *scale) {
		
	PyTransform *t = PyObject_New(PyTransform, &PyTransformType);

        D_ALLOC(t, "Transform");
        
	t->interfaces = &PyTransform_Interfaces;
		
	if (translate == NULL) {
		translate = C_PyVector_FromValues(0, 0, 0);
	}
	Py_INCREF(translate);
	
	if (rotateAxis == NULL) {
		rotateAxis = C_PyVector_FromValues(0, 0, 1);
		rotateAmount = 0;
	}
	Py_INCREF(rotateAxis);
	
	if (scale == NULL) {
		scale = C_PyVector_FromValues(1, 1, 1);
	}
	Py_INCREF(scale);
		
	t->translation = translate;
	t->rotationAxis = rotateAxis;
	t->rotationAmount = rotateAmount;
	t->scale = scale;
	
	return (PyObject *)t;
	
}

PyObject *PyTransform_New(PyObject *self, PyObject *args, PyObject *kwords) {

	PyObject *translate = NULL, *rotate = NULL, *scale = NULL;
	double amount;
	
        static char *kwlist[] = {"translate", "rotateAxis", 
                "rotateAmount", "scale", NULL};
	
	if (!PyArg_ParseTupleAndKeywords(args, kwords, "|OOdO", kwlist, &translate, 
		&rotate, &amount, &scale)) {
		return NULL;
	}
	
	return C_PyTransform_New(translate, rotate, amount, scale);
	
}

static void C_PyTransform_Free(PyObject *self) {
	
	PyTransform *t = (PyTransform *)self;

        D_FREE(t, "Transform");
        
	Py_DECREF(t->translation);
	Py_DECREF(t->rotationAxis);
	Py_DECREF(t->scale);
	
	PyObject_Del(self);
	
}

static void PyTransform_Free(PyObject *self) {
	
	C_PyTransform_Free(self);
	
}

static void C_PyTransform_SetTranslation(PyObject *self, PyObject *translation) {
	
	PyTransform *t = (PyTransform *)self;
	
	Py_DECREF(t->translation);
	Py_INCREF(translation);
	t->translation = translation;
	
}

static PyObject *PyTransform_SetTranslation(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}
	
	C_PyTransform_SetTranslation(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static void C_PyTransform_SetRotationAxis(PyObject *self, PyObject *axis) {
	
	PyTransform *t = (PyTransform *)self;
	
	Py_DECREF(t->rotationAxis);
	if (axis != NULL)
		Py_INCREF(axis);
	t->rotationAxis = axis;
	
}

static PyObject *C_PyTransform_GetRotationAxis(PyObject *self) {

	PyTransform *t = (PyTransform *)self;
	
	if (t->rotationAxis != NULL)
	{
		Py_INCREF(t->rotationAxis);
	}
	
	return t->rotationAxis;
}

static PyObject *PyTransform_GetRotationAxis(PyObject *self, PyObject *args) {

	PyObject *ret = C_PyTransform_GetRotationAxis(self);

	if (ret == NULL) {
		Py_INCREF(Py_None);
		return Py_None;
	} else {
		return ret;
	}
}
	
static PyObject *PyTransform_SetRotationAxis(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}

	if (arg == Py_None)
		arg = NULL;
	
	C_PyTransform_SetRotationAxis(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static void C_PyTransform_SetRotationAmount(PyObject *self, double amount) {
	
	PyTransform *t = (PyTransform *)self;
	
	t->rotationAmount = amount;
	
}

static double C_PyTransform_GetRotationAmount(PyObject *self) {

	PyTransform *t = (PyTransform *)self;

	return t->rotationAmount;

}

static PyObject *PyTransform_SetRotationAmount(PyObject *self, PyObject *args) {
	
	double val;
	if (!PyArg_ParseTuple(args, "d", &val)) {
		return NULL;
	}
	
	C_PyTransform_SetRotationAmount(self, val);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static PyObject *PyTransform_GetRotationAmount(PyObject *self, PyObject *args) {

	double val = C_PyTransform_GetRotationAmount(self);

	return PyFloat_FromDouble(val);

}

static void C_PyTransform_SetScale(PyObject *self, PyObject *scale) {
	
	PyTransform *t = (PyTransform *)self;
	
	Py_DECREF(t->scale);
	Py_INCREF(scale);
	t->scale = scale;
	
}

static PyObject *PyTransform_SetScale(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}
	
	C_PyTransform_SetScale(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

inline void C_PyTransform_TransformByMe(PyObject *self) {

	PyTransform *t = (PyTransform *)self;
        PyVector *v = (PyVector *)t->translation;

        //printf("translate: (%.2f,%.2f,%.2f)\n", v->vector[0], v->vector[1],
        //                v->vector[2]);
	
	glTranslated(v->vector[0], v->vector[1], v->vector[2]);
        v = (PyVector *)t->scale;	
	
	if (v->vector[0] != 0)
	{
#if DEBUG_VERBOSE
		printf("scale: (%.2f,%.2f,%.2f)\n", v->vector[0], v->vector[1],
        	                v->vector[2]);
#endif

	        glScaled(v->vector[0], v->vector[1], v->vector[2]);
	}
        if (t->rotationAmount != 0)
        {
                v = (PyVector *)t->rotationAxis;
#if DEBUG_VERBOSE		
                printf("rotate: %.2f @ (%.2f,%.2f,%.2f)\n", t->rotationAmount, 
                                v->vector[0], v->vector[1], v->vector[2]);
#endif
	        glRotated(t->rotationAmount, v->vector[0], v->vector[1], 
                                v->vector[2]);
        }
	
	
}

inline void C_PyTransform_InvertByMe(PyObject *self) {

        PyTransform *t = (PyTransform *)self;
        PyVector *v;

        if (t->rotationAmount != 0)
        {
                v = (PyVector *)t->rotationAxis;
                glRotated(-t->rotationAmount, v->vector[0], v->vector[1],
                                v->vector[2]);
        }

        v = (PyVector *)t->scale;
        glScaled(1/v->vector[0], 1/v->vector[1], 1/v->vector[2]);

        v = (PyVector *)t->translation;
        glTranslated(-v->vector[0], -v->vector[1], -v->vector[2]);

}

static void C_PyTransform_GenerateTransform(PyObject *self, PyObject *status,
                int time, int gen_forward, int gen_reverse, int apply)
{

        if (apply == 1)
                C_PyTransform_TransformByMe(self);
        else if (apply == -1)
                C_PyTransform_InvertByMe(self);

}
        

static PyMethodDef PyTransform_methods[] = {
        {"setTranslation", (PyCFunction)PyTransform_SetTranslation,METH_VARARGS,
                "Sets the translation vector of this transform"},
        {"setRotationAxis", (PyCFunction)PyTransform_SetRotationAxis, 
                METH_VARARGS, "Set the rotation axis of this transform"},
        {"getRotationAxis", (PyCFunction)PyTransform_GetRotationAxis, 
                METH_VARARGS, "Get the rotation axis of this transform"},
        {"setRotationAmount", (PyCFunction)PyTransform_SetRotationAmount,
                METH_VARARGS, "Set the angle of rotation for this transform"},
        {"getRotationAmount", (PyCFunction)PyTransform_GetRotationAmount,
                METH_VARARGS, "Get the angle of rotation for this transform"},
        {"setScale", (PyCFunction)PyTransform_SetScale, METH_VARARGS,
                "Sets the scale vector of this transform"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyTransform_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyTransform_methods, self, name);

}

static PyTypeObject PyTransformType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyTransform",
        sizeof(PyTransform),
        0,
        PyTransform_Free,
        0, /* print */
        PyTransform_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
