#include "TograIncludes.h"

staticforward PyTypeObject PyTransformContainerType;

/*
typedef struct {
        void (*generateTransform)(PyObject *self, PyObject *status, int time,
                        int gen_forward, int gen_reverse, int apply);
        void (*transformByMe)(PyObject *self);
        void (*invertByMe)(PyObject *self);
        int (*registerSize)(PyObject *self);
        void (*registerFunc)(PyObject *self, PyObject *status, int delay,
                        PyObject *callback, PyObject *user, PyObject *user2);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
} ITransform;
*/

int PyTransformContainer_Check(PyObject *v) {

        return v->ob_type == &PyTransformContainerType;

}

void C_PyTransformContainer_GenerateTransform(PyObject *self, PyObject *status,
                int time, int gen_forward, int gen_reverse, int apply) {

        PyTransformContainer *tc = (PyTransformContainer *)self;
        int i;
#if DEBUG_VERBOSE
	printf("Called transformContainer Generate method on %p\n",
			self);
#endif
        
	if (apply == 1) {
                for (i = tc->length - 1; i >= 0; i--)
                        GENERATE(tc->children[i], tc->status[i], time,
                                        gen_forward, gen_reverse, apply);
        } else {
                for (i = 0; i < tc->length; i++)
                        GENERATE(tc->children[i], tc->status[i], time,
                                        gen_forward, gen_reverse, apply);
        }
}

void C_PyTransformContainer_TransformByMe(PyObject *self) {

        PyTransformContainer *tc = (PyTransformContainer *)self;
        int i;

#if DEBUG_VERBOSE
	printf("Called transformContainer TransformByMe method on %p\n",
			self);
#endif
	
        for (i = tc->length - 1; i >= 0; i--)
                TRANSFORMBYME(tc->children[i]);
}

void C_PyTransformContainer_InvertByMe(PyObject *self) {

        PyTransformContainer *tc = (PyTransformContainer *)self;
        int i;

        for (i = 0; i < tc->length; i++)
                INVERTBYME(tc->children[i]);
}

int C_PyTransformContainer_RegisterSize(PyObject *self) {return 0;}

static ITransform PyTransformContainer_ITransform = {
        C_PyTransformContainer_GenerateTransform,
        C_PyTransformContainer_TransformByMe,
        C_PyTransformContainer_InvertByMe,
        C_PyTransformContainer_RegisterSize
};

static IC PyTransformContainer_Interfaces = {NULL, NULL, NULL, NULL, NULL,
        &PyTransformContainer_ITransform};

PyObject *C_PyTransformContainer_New(void) {

        PyTransformContainer *c = (PyTransformContainer *)
                PyObject_New(PyTransformContainer, &PyTransformContainerType);

        D_ALLOC(c, "TransformContainer");

        c->interfaces = &PyTransformContainer_Interfaces;
        c->alloced = c->length = 0;
        c->children = NULL;
        c->status = NULL;

        return (PyObject *)c;
}

PyObject *C_PyTransformContainer_OfLength(int length) {

        PyTransformContainer *c
                = (PyTransformContainer *)C_PyTransformContainer_New();
        c->alloced = length;
        c->children = malloc(sizeof(PyObject *) * length);
        c->status = malloc(sizeof(PyObject *) * length);

        return (PyObject *)c;

}

PyObject *PyTransformContainer_New(PyObject *self, PyObject *args) {

        int length;
        PyObject *o;


        if (PyTuple_Size(args) == 0) {
                return (PyObject *)C_PyTransformContainer_New();
        }

        if (!PyArg_ParseTuple(args, "i", &length)) {
                return NULL;
        }

        if (length == 0)
                return (PyObject *)C_PyTransformContainer_New();

        o = (PyObject *)C_PyTransformContainer_OfLength(length);
        return o;

}

static void C_PyTransformContainer_Free(PyObject *self) {

        PyTransformContainer *c = (PyTransformContainer *)self;
        int i;


        D_FREE(c, "TransformContainer");
        
        if (c->alloced > 0) {

                for (i = 0; i < c->length; i++) {
                        Py_DECREF(c->children[i]);
                }

                free(c->children);
        }

        if (!(c->status == NULL))
                free(c->status);

        PyObject_Del((PyObject *)c);

}

static void PyTransformContainer_Free(PyObject *self) {

        C_PyTransformContainer_Free(self);

}

void *C_PyTransformContainer_Append(PyObject *self, PyObject *object,
                PyObject *args) {

        PyTransformContainer *c = (PyTransformContainer *)self;

        int size;
	void *r;

        Py_INCREF(object);

        if (c->length == c->alloced) {
                if (c->alloced == 0) {
                        c->alloced++;
                }
                else {
                        c->alloced *= 2;
                }

                c->children = realloc(c->children,
                                c->alloced * sizeof(PyObject *));
                c->status = realloc(c->status,
                                c->alloced * sizeof(PyObject *));
        }

        c->children[c->length] = object;

        size = REGISTERSIZE(object);

        if (size == 0)
        {
                c->status[c->length] = NULL;
		r = NULL;
        } else {
                c->status[c->length] = malloc(size);
                r = REGISTER(object, c->status[c->length], args);
        }

        c->length += 1;

	return r;
}

static PyObject *PyTransformContainer_Append(PyObject *self, PyObject *args) {

        PyObject *innards;
        PyObject *object, *arg = NULL;
	void *r;

        if (!PyArg_ParseTuple(args, "O|O", &innards, &arg)) {
                return NULL;
        }

        if (PyTuple_Check(innards)) {
                if (!PyArg_ParseTuple(innards, "O|O", &object, &arg)) {
                        return NULL;
                }
        } else {
                object = innards;
        }

        r = C_PyTransformContainer_Append(self, object, arg);

	return PyInt_FromLong((long)r);

}
PyObject *PyTransformContainer_GetItemHandle(PyObject *self, PyObject *args) {

	int pos = 0;
        PyTransformContainer *c = (PyTransformContainer *)self;
	
	if (!PyArg_ParseTuple(args, "i", &pos)) {
		return NULL;
	}

	return PyInt_FromLong((long)c->status[pos]);

}


static PyMethodDef PyTransformContainer_methods[] = {
        {"append", (PyCFunction)PyTransformContainer_Append, METH_VARARGS,
                "Appends an object to the container"},
	{"getItemHandle", (PyCFunction)PyTransformContainer_GetItemHandle,
		METH_VARARGS, "Get's a TC child's handle"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyTransformContainer_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyTransformContainer_methods, self, name);

}

static PyObject *PyTransformContainer_Item(PyObject *self, int index) {

        PyTransformContainer *c = (PyTransformContainer *)self;

        if (index < c->length && index >= 0) {
		Py_INCREF(c->children[index]);
		return c->children[index];
        }

        return NULL;

}

static int PyTransformContainer_AssItem(PyObject *self, int index, PyObject *item) {

        PyTransformContainer *c = (PyTransformContainer *)self;

        PyObject *obj = item;
        PyObject *args = NULL;

        int size;

        if (PyTuple_Check(item) && PyTuple_Size(item) > 1) {
                obj = PyTuple_GetItem(item, 0);
                args = PyTuple_GetItem(item, 1);
        }

        if (index < c->length && index >= 0) {
                if (c->status[index] != NULL) {
                    DEREGISTER(c->children[index], c->status[index]);
                    free(c->status[index]);
                }
                Py_DECREF(c->children[index]);

                Py_INCREF(obj);
                size = REGISTERSIZE(obj);

                if (size > 0) {
                        c->status[index] = malloc(size);
                        REGISTER(obj, c->status[index], args);
                } else {
                        c->status[index] = NULL;
                }

                c->children[index] = obj;

                return 0;
        }

        return 1;

}

static PySequenceMethods PyTransformContainer_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        PyTransformContainer_Item, /* item */
        0, /* slice */
        PyTransformContainer_AssItem, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyTransformContainerType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyContainer",
        sizeof(PyTransformContainer),
        0,
        PyTransformContainer_Free,
        0, /* print */
        PyTransformContainer_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyTransformContainer_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
