from Togra import *
from TograUtil import *
import math
from Graph import *
from OpenGL.GL import *
#import Tracker
import random
import sys

class ProtoGraphMixin:
	def __init__(this, graph):
		this.pm = ProtoMixin()
		this.pm.setInit(this.f_init)
		this.pm.setRenderStep(this.f_rs)
		this.graph = graph
		this.schemas = []
		this.schema_maps = []

	def f_init(this, time):
		pass

	def f_rs(this, time, is_paused, paused_time):
		i = 0
		for schema in this.schemas:
			if schema[0] == "ALL":
				this.render_ALL(schema[1:])
			elif schema[0] == "ALL_CONNECTED":
				this.render_ALL_CONNECTED(schema[1:])
			elif schema[0] == "ALL_PAIRS":
				this.render_ALL_PAIRS(schema[1:], i)
			elif schema[0] == "ALL_GROUPS":
				this.render_ALL_GROUPS(schema[1:])
			elif schema[0] == "ALL_PAIRS_WITHIN_GROUP":
				this.render_ALL_PAIRS_WITHIN_GROUP(schema[1:])
			else:
				print "Bad schema type: %s" % (schema[0],)
			i += 1

	def add_schema(this, data):
		this.schemas.append(data)
		this.schema_maps.append(None)
	
	def render_ALL_PAIRS_WITHIN_GROUP(this, data):
		if data[0] == "INVERSE":
			func = lambda d: data[1]/d.length()/d.length()
		elif data[0] == "INVERSE_SQUARE":
			func = lambda d: data[1]/math.pow(d.length(),3)
		elif data[0] == "LINEAR_OFFSET":
			func = lambda d:data[2]*(1 - data[1]/d.length()) 
		else:
			print "invalid ALL_WITHIN_GROUP force type"
			return
		
		tt = this.graph.StartGetGroups()
		while 1:
			group = this.graph.NextGroup(tt)
			if group is None:
				break

			t1 = this.graph.StartGetGroupNodes(group)
			while 1:
				f = this.graph.GetNextGroupNode(group, t1)
				if f is None:
					break
				f = f.transformContainer[1]
				t2 = this.graph.StartGetGroupNodes(group, t1)
				while 1:
					t =this.graph.GetNextGroupNode(group,t2)
					if t is None:
						break
					t = t.transformContainer[1]
					
					fp = f.getPosition()
					tp = t.getPosition()
					d = tp - fp
					fo = func(d)
					t.addForce(Vector(
						d[0]*fo,d[1]*fo,d[2]*fo))
					fo = -fo
					f.addForce(Vector(
						d[0]*fo,d[1]*fo,d[2]*fo))

				
		

	def render_ALL_GROUPS(this, data):
		if data[0] == "INVERSE":
			func = lambda d: data[1]/d.length()/d.length()
		elif data[0] == "INVERSE_SQUARE":
			func = lambda d: data[1]/math.pow(d.length(),3)
		elif data[0] == "LINEAR_OFFSET":
			func = lambda d:data[2]*(1 - data[1]/d.length()) 
		else:
			print "invalid ALL_GROUPS force type"
			return
		t1 = this.graph.StartGetGroups()
		all_group_nodes = []
		group_cogs = []
		while 1:
			f = this.graph.NextGroup(t1)
			if f is None:
				break
			fnodes = []
			fnt = this.graph.StartGetGroupNodes(f)
			p = Vector(0,0,0)
			n = 0
			while 1:
				tmp = this.graph.GetNextGroupNode(f, fnt)
				if tmp is None:
					break
				tmp = tmp.transformContainer[1]
				fnodes.append(tmp)
				p += tmp.getPosition()
				n += 1
			
			all_group_nodes.append(fnodes)
			group_cogs.append(Vector(p[0]/n, p[1]/n, p[2]/n))

		for i in range(len(all_group_nodes)):
			for j in range(i+1, len(all_group_nodes)):
				f = group_cogs[i]
				t = group_cogs[j]
				d = t - f
				force = func(d)
				ff = force * len(all_group_nodes[j])
				for node in all_group_nodes[i]:
					node.addForce(Vector(ff*d[0], ff*d[1],
						ff*d[2]))
				ff = -force * len(all_group_nodes[i])
				for node in all_group_nodes[j]:
					node.addForce(Vector(ff*d[0], ff*d[1],
						ff*d[2]))

	def render_ALL(this, data):
		if data[0] == "CONSTANT":
			func = lambda d:Vector(data[1]*d[0], data[1]*d[1],\
				data[1]*d[2])
		elif data[0] == "TO_Z_PLANE":
			func = lambda d:Vector(0,0,-data[1]*d[2])
		t = this.graph.StartGetNodes()
		if data[0] == "CONSTANT":
			while 1:
				node = this.graph.GetNextNode(t)
				if node is None:
					break
				node = node.transformContainer[1]
				
				d = node.getVelocity()
				force = func(d)
				node.addForce(force)
		elif data[0] == "TO_Z_PLANE":
			while 1:
				node = this.graph.GetNextNode(t)
				if node is None:
					break
				node = node.transformContainer[1]
				
				d = node.getPosition()
				force = func(d)
				node.addForce(force)



	def render_ALL_CONNECTED(this, data):
		if data[0] == "SQUARE_OFFSET":
			func = lambda d:data[2]*(d.length() - data[1]) 
		elif data[0] == "LINEAR_OFFSET":
			func = lambda d:data[2]*(1 - data[1]/d.length()) 
		else:
			print "invalid ALL_CONNECTED force type"
			return

		token = this.graph.StartGetConnections()
		while 1:
			con = this.graph.GetNextConnection(token)
			if con is None:
				break
			f = con[0].transformContainer[1]
			t = con[1].transformContainer[1]
			fpos = f.getPosition()
			tpos = t.getPosition()
			d = tpos - fpos
			force = func(d)

			f.addForce(Vector(d[0]*force, d[1]*force, d[2]*force))
			force = -force
			t.addForce(Vector(d[0]*force, d[1]*force, d[2]*force))

	def render_ALL_PAIRS(this, data, index):
		if data[0] == "INVERSE":
			func = lambda d: data[1]/d.length()/d.length()
		elif data[0] == "INVERSE_SQUARE":
			func = lambda d: data[1]/math.pow(d.length(),3)
		else:
			print "invalid ALL_PAIRS force type"

		schema_map = this.schema_maps[index]
		if schema_map == None and data[-1] > 0:
			g = data[-1]
			schema_map = [{}, {}, g]
			t1 = this.graph.StartGetNodes()
			while 1:
				node = this.graph.GetNextNode(t1)
				if node is None:
					break
				node = node.transformContainer[1]
				node_pos = node.getPosition()
				node_loc = (int(node_pos[0])/g,
					int(node_pos[1])/g, int(node_pos[2])/g)
				schema_map[0][node] = node_loc
				try:
					schema_map[1][node_loc].append(node)
				except:
					schema_map[1][node_loc] = [node]
			this.schema_maps[index] = schema_map

		if schema_map is not None:
			g = schema_map[2]
			for node in schema_map[0].keys():
				node_pos = node.getPosition()
				node_loc = (int(node_pos[0])/g,
					int(node_pos[1])/g, int(node_pos[2])/g)
				if not node_loc == schema_map[0][node]:
					old_loc = schema_map[0][node]
					schema_map[0][node] = node_loc
					i = schema_map[1][old_loc].index(node)
					del schema_map[1][old_loc][i]
					if len(schema_map[1][old_loc]) == 0:
						del schema_map[1][old_loc]
					try:
						schema_map[1][node_loc].append(\
							node)
					except:
						schema_map[1][node_loc] = \
							[node]
			#for key in schema_map[1].keys():
			#	print key, "->", schema_map[1][key]
			
			nlists = 0
			nignored = 0
			npairs = 0
			for loc in schema_map[1].keys():
				list = []
				for i in [-1,0,1]:
				  for j in [-1,0,1]:
				    for k in [-1,0,1]:
				      tp = (loc[0]+i, loc[1]+j, loc[2]+k)
				      try:
				        list += schema_map[1][tp]
				      except:
				        pass

				if len(list) < 2:
					nignored += len(list)
					continue
				nlists += 1	
				for i in range(len(schema_map[1][loc])):
					f = schema_map[1][loc][i]
					fpos = f.getPosition()
					for j in range(i+1, len(list)): 
						t = list[j]
						if t == f:
							continue
						tpos = t.getPosition()

						d = tpos - fpos
						if d.length() > g:
							continue
						npairs += 1
						force = func(d)
						t.addForce(Vector(d[0]*force, 
							d[1]*force, d[2]*force))
						force = -force
						f.addForce(Vector(d[0]*force, 
							d[1]*force, d[2]*force))
						
			#print "%d lists calced, %d ignored, %d pair calcs required" % (nlists, nignored, npairs)
			return		

		t1 = this.graph.StartGetNodes()
		while 1:
			f = this.graph.GetNextNode(t1)
			if f is None:
				break
			f = f.transformContainer[1]
			t2 = this.graph.StartGetNodes(t1)
			while 1:
				t = this.graph.GetNextNode(t2)
				if t is None:
					break
				t = t.transformContainer[1]
				fpos = f.getPosition()
				tpos = t.getPosition()
				
				d = tpos - fpos
				force = func(d)
				t.addForce(Vector(d[0]*force, d[1]*force,
					d[2]*force))
				force = -force
				f.addForce(Vector(d[0]*force, d[1]*force,
					d[2]*force))
	
class ProtoGraphRender:
	def __init__(this, graph):
		this.pr = ProtoRender()
		this.pr.setRender(this.render)
		this.pr.setCompile(this.compile)
		this.graph = graph
	
	def compile(this):
		return 0

	def render(this, time):
		glBegin(GL_LINES)
		token = this.graph.StartGetConnections()
		while 1:
			c = this.graph.GetNextConnection(token)
			if c is None:
				break
			t = c[0].transformContainer[1].getPosition()
			f = c[1].transformContainer[1].getPosition()
			glVertexd(t[0],t[1],t[2])
			glVertexd(f[0],f[1],f[2])
		glEnd()

		token = this.graph.StartGetNodes()
		while 1:
			n = this.graph.GetNextNode(token)
			if n is None:
				break
			n()

scene = init()
graph = Graph()
pgm = ProtoGraphMixin(graph)
#pgm.add_schema(["ALL", "CONSTANT", 10.0, 0, "FROM", Vector(1,0,0)])
#pgm.add_schema(["ALL", "LINEAR", 8.0, 40.0, 0, "FROM", Vector(0,0,0)])
#pgm.add_schema(["ALL", "CONSTANT", 10.0, 0, "CONSTANT", Vector(0,0,1)])
pgm.add_schema(["ALL_CONNECTED", "LINEAR_OFFSET", 12.0, 20.0, 0])
#pgm.add_schema(["ALL", "INVERSE_SQUARE", 10.0, 0, "FROM", Vector(0,0,0)])
pgm.add_schema(["ALL", "CONSTANT", -10.0, 0, "WITH_VELOCITY"])
pgm.add_schema(["ALL_PAIRS", "INVERSE_SQUARE", 500.0, 0])
#pgm.add_schema(["ALL_GROUPS", "LINEAR_OFFSET", 20.0, 2.0, 0])
#pgm.add_schema(["ALL_PAIRS_WITHIN_GROUP", "INVERSE_SQUARE", 500.0, 0])

ode = ODEMixin()

scene.view_from = Vector(0,0,-200)
#t = Tracker.initTracker(scene)

is_pressed = 0
last_x = 0
last_y = 0
zoom = 0
pan = 1
select = 2
mode = pan
oldsel = []
msel = Material(ambient=(1,0,0,1))
munsel = Material(ambient=(0.5,0.5,0.5,1))
def mouseFunc(*args):
	global is_pressed, last_x, last_y, mode, oldsel
	if args[0] == 2:
		mode = zoom
	else:
		mode = pan
	is_pressed = args[1]
	if is_pressed:
		last_x = args[2]
		last_y = args[3]
		l = scene.pick(args[2], args[3])
		for item in oldsel:
			item.setMaterial(munsel)
		oldsel = []
		for item in l:
			if item[2] is not None:
				if type(item[2]) is not type(Container()):
					item[2].setMaterial(msel)
					oldsel.append(item[2])
					mode = select

def motionFunc(*args):
	global is_pressed, last_x, last_y, mode
	if not is_pressed:
		return
	difx = args[0] - last_x
	dify = args[1] - last_y
	last_x = args[0]
	last_y = args[1]
	ratio = scene.view_from[2] / -1000.0
	if mode == pan:
		scene.view_from += Vector(difx*ratio, dify*ratio, 0)
		scene.view_to += Vector(difx*ratio, dify*ratio, 0)
	elif mode == select:
		for item in oldsel:
			g = item.transformContainer[1]
			p = g.getPosition()
			p -= Vector(difx*ratio, dify*ratio, 0)
			g.setPosition(p)
	else:
		scene.view_from += Vector(0, 0, dify)
		last_y = args[1]

scene.mouseFunction = mouseFunc
scene.motionFunction = motionFunc

#scene.baseContainer.transformContainer.append(t)

scene.addMixin(pgm.pm)
scene.addMixin(ode)
ode.setGravity(Vector(0,0,0))


def other(key):
	if key == ord('l'):
		pgm.schemas[2][-1] = 50
	elif key == ord('g'):
		pgm.schemas[2][-1] = 0
		pgm.schema_maps[2] = None
	elif key == ord('3'):
		token2 = graph.StartGetNodes()
		while 1:
			node = graph.GetNextNode(token2)
			if node is None:
				break
			n = node.transformContainer[1]
			v = n.getVelocity()
			v += Vector(0,0,random.random()*200-100)
			n.setVelocity(v)
	elif key == ord('2'):
		pgm.add_schema(["ALL", "TO_Z_PLANE", 50.0])
		print "added"

pauseQuitKeystrokes(scene, other)

def add_node(position, velocity):
	#velocity = Vector(
	#	random.random()*30 - 15,
	#	random.random()*30 - 15,
	#	random.random()*30 - 15)
	global ode, graph, scene
	t = sphere(0, 0, 0, 1, 30, 30)
	t.setMaterial(munsel)
	o = ODETransform(ode)
	o.setSphere(1, 1)
	o.setPosition(position)
	o.setVelocity(velocity)
	t.transformContainer.append(o)
	r = graph.AddNode(t)
	#scene.baseContainer.append(t)
	return r

def add_connection(a, b):
	global graph
	graph.AddConnection(a, b)

scene.newLight(Vector(-10,10,0), (0.8,0.2,0.2,1))
scene.newLight(Vector(20,0,-10), (0.2,0.2,0.8,1))

if len(sys.argv) == 3:
	size = int(sys.argv[1])
	connects = int(sys.argv[2])
	nodes = []
	for i in range(size):
		nodes.append(add_node(Vector(0,0,0), Vector(0,0,0)))
	for i in range(connects):
		ca = random.randint(0, size - 1)
		cb = random.randint(0, size - 2)
		if ca <= cb:
			cb += 1
		add_connection(nodes[ca], nodes[cb])

	token = graph.StartGetGroups()
	while 1:
		group = graph.NextGroup(token)
		if group is None:
			break
		groupPosX = random.random() * 100 - 50
		groupPosY = random.random() * 100 - 50

		token2 = graph.StartGetGroupNodes(group)
		while 1:
			node = graph.GetNextGroupNode(group, token2)
			if node is None:
				break
		
			x = random.random() * 30 - 15 + groupPosX
			y = random.random() * 30 - 15 + groupPosY
			node.transformContainer[1].setPosition(Vector(x,y,0))

	
	
else:
	a = add_node(Vector(0,0,2), Vector(0,0,0))
	b = add_node(Vector(0,0,0), Vector(0,0,0))
	c = add_node(Vector(0,2,0), Vector(0,0,0))
	d = add_node(Vector(2,0,0), Vector(0,0,0))
	e = add_node(Vector(5,0,0), Vector(0,0,0))
	f = add_node(Vector(5,5,0), Vector(0,0,0))

	add_connection(a, d)
	add_connection(b, d)
	add_connection(c, d)
	add_connection(a, c)
	add_connection(b, c)
	add_connection(e,f)

pgr = ProtoGraphRender(graph)
scene.baseContainer.append(pgr.pr)

scene.run()
