from Togra import *
import math

GL_QUADS=7
LINEAR_PATH=1

def length(vector):
        return math.sqrt(vector * vector)

def normalise(vector):
        size = length(vector)
        return Vector(vector[0] / size, vector[1] / size, vector[2] / size)

def times(vector, double):
        return Vector(vector[0] * double, vector[1] * double, 
                vector[2] * double)

def circlePath(axis, start, steps, time):

        ip = InterpolatedPath(steps + 1)
        diff = start
        vx = diff[1] * axis[2] - diff[2] * axis[1]
        vy = diff[2] * axis[0] - diff[0] * axis[2]
        vz = diff[0] * axis[1] - diff[1] * axis[0]
        
        diff2 = normalise(Vector(vx, vy, vz))
        diff2 = times(diff2, length(diff))
       
        for i in range(steps + 1):
                angle = i * 2 * math.pi / steps
                pos = times(diff, math.cos(angle)) + \
                        times(diff2, math.sin(angle))
                pos90 = times(diff, math.cos(angle - math.pi/2)) + \
                        times(diff2, math.sin(angle - math.pi/2))
                posn90 = times(pos90, -1)
                ip.append((pos, times(pos90, 2.0 / steps), 
                        times(posn90, 2.0 / steps)), i * time / steps)
        ip.setMaxIndex(time)

        return ip

def fromToPath(start, end, time):
        ip = InterpolatedPath(2)
        ip.setType(LINEAR_PATH)
        ip.append(start, 0)
        ip.append(end, time)
        ip.setMaxIndex(time)

        return ip

def rotatePath(axis, period):
        ip = InterpolatedPath(2)
        ip.setType(2)
        ip.append((axis, 0), 0)
        ip.append((axis, 360), period)
        ip.setMaxIndex(period)
        return ip

_ISPAUSED = [0]
_world = None
_if_not = None

def _pause_and_quit(*args):
        global _world, _if_not
        if args[0] == ord('q'):
                _world.exit()                # this will exit!
        elif args[0] == ord(' '):
                if _ISPAUSED[0]:
                       _world.cont()
                else:
                       _world.pause()
                _ISPAUSED[0] = not _ISPAUSED[0]
	elif _if_not is not None:
		_if_not(args[0])

def pauseQuitKeystrokes(world, if_not = None):
        global _world, _if_not
        world.keyFunction = _pause_and_quit
        _world = world
	_if_not = if_not

def quadFace(x0, y0, x1, y1, x2, y2, x3, y3):
	rbox = VectorList(4)
	rbox.append(Vector(x0, y0, 0))
	rbox.append(Vector(x1, y1, 0))
	rbox.append(Vector(x2, y2, 0))
	rbox.append(Vector(x3, y3, 0))

	face = Container(1)
	sr = StaticRender(rbox, GL_QUADS)
	face.append(sr)
	
	return face

def from_polar(v):
	vr = [0,0,0]
	vr[1] = math.sin(v[1])
	# v0/v2 = x
	# v0^2 + v2^2 = y
	# v2^2x^2 + v2^2 = y
	# v2 = math.sqrt(y / (x^2 + 1))
	ratio = math.tan(v[0])
	size = 1 - vr[1]*vr[1]
	vr[2] = math.sqrt(size / (ratio * ratio + 1))
	if (math.cos(v[0]) <= 0 or math.cos(v[1]) <= 0):
		vr[2] = -vr[2]
	vr[0] = vr[2] * ratio
	return vr

def to_polar(v):
	if (v[2] == 0):
		if v[0] > 0:
			r = math.pi/2
		else:
			r = -math.pi/2 
	else:
		r = math.atan(v[0]/v[2])
		if (v[2] < 0):
			r += math.pi
	h = math.asin(v[1])
	return (r, h)


#ip = InterpolatedPath(5)
#ip.append((Vector(30,0,0), Vector(0,-15,0), Vector(0,15,0)),0)
#ip.append((Vector(0,30,0), Vector(15,0,0), Vector(-15,0,0)), 1000)
#ip.append((Vector(-30,0,0), Vector(0,15,0), Vector(0,-15,0)), 2000)
#ip.append((Vector(0,-30,0), Vector(-15,0,0), Vector(15,0,0)),3000)
#ip.append((Vector(30,0,0), Vector(0,-15,0), Vector(0,15,0)),4000)
#ip.setMaxIndex(4000)
